// Beltrani Saele 0000311983


import java.io.*;
import java.net.*;
import java.util.StringTokenizer;

public class MulticastServer {
// uso: MulticastServer portDiscoveryServer ipMCast portMCast file

    public static void main(String[] args) {
        /*
         * protocollo:
         * chiedo connessione con "MC":IP:PORT:FILE
         * ottengo "si" o "no" se è accettata o meno la connessione
         */
        //test
        args = new String[4];
        args[0] = "65535";
        args[1] = "226.0.0.1";
        args[2] = "5000";
        args[3] = "saggezza.txt";

        // controllo argomenti, inizializzazione, registrazione
        File f = null;
        DatagramSocket discoverySocket = null;
        MulticastSocket multicastSock = null;
        DatagramPacket pack = null;
        InetAddress localAddr = null;
        InetAddress multicastAddr = null;
        int discoveryPort = -1;
        int multicastPort = -1;
        String richiesta;
        String risposta;
        byte[] buf = new byte[256];

        try {
            discoverySocket = new DatagramSocket();
            discoverySocket.setSoTimeout(10000);
            localAddr = InetAddress.getLocalHost();

            // richiesta porta DiscoveryServer
            Boolean rightPort = false;
            while(!rightPort){
                System.out.println("Scrivi la porta del Discovery al quale connetterti");
                BufferedReader br = new BufferedReader(new InputStreamReader(System.in));
                try {
                    discoveryPort = Integer.parseInt(br.readLine());
                    if(discoveryPort < 1 || discoveryPort > 65535) throw new NumberFormatException();
                } catch(NumberFormatException nfex){
                    System.out.println("Porta non valida");
                    continue;
                } //catch
                rightPort = true;
            } //while
            pack = new DatagramPacket(buf, buf.length, localAddr, discoveryPort);

            // richiesta indirizzo e porta di multicast
            Boolean rightEndpoint = false;
            while(!rightEndpoint){
                System.out.println("Immetti indirizzo di multicast e porta in forma IP:PORT");
                BufferedReader br = new BufferedReader(new InputStreamReader(System.in));
                try{
                    StringTokenizer str = new StringTokenizer(br.readLine(), ":");
                    if(str.countTokens() != 2){
                        throw new StringIndexOutOfBoundsException();
                    } //if
                    multicastAddr = InetAddress.getByName(str.nextToken());
                    if(!multicastAddr.isMulticastAddress()) throw new UnknownHostException("NOMC");
                    multicastPort = Integer.parseInt(str.nextToken());
                    if(multicastPort < 1 || multicastPort > 65535) throw new NumberFormatException();
                } catch(NumberFormatException nfex){
                    System.out.println("Porta non valida");
                    continue;
                } catch(UnknownHostException uhex){
                    if(uhex.getMessage().equals("NOMC")){
                        System.out.println("L'indirizzo non è di multicast");
                        continue;
                    } else{
                        System.out.println("Indirizzo non valido");
                        continue;
                    } //if-else
                } catch(StringIndexOutOfBoundsException stex){
                    System.out.println("Forma errata, usa IP:PORTA");
                    continue;
                } //try-catch
                rightEndpoint = true;
            } //while

            //richiesta file
            Boolean rightFile = false;
            while(!rightFile){
                System.out.println("Scrivi il file da cui leggere");
                BufferedReader br = new BufferedReader(new InputStreamReader(System.in));
                f = new File(br.readLine());
                if (!f.exists()) {
                    System.out.println("File inesistente");
                    continue;
                }
                rightFile = true;
            } //while

            //mando la richiesta di connessione
            richiesta = "MC:" + multicastAddr.getHostAddress() + ":" + multicastPort + ":" + f.getPath();
            pack = Utility.insertStringInPacket(pack, richiesta);
            discoverySocket.send(pack);

            //attendo la risposta del discovery
            buf = new byte[256];
            pack.setData(buf);
            discoverySocket.receive(pack);
            risposta = Utility.getStringFromPacket(pack);

            //valuto la risposta
            if (risposta.equals("no")) {
                System.out.println("Richiesta di sottoscrizione non accettata");
                System.exit(1);
            } else if (risposta.equals("si")) {

                // apro la socket di multicast
                multicastSock = new MulticastSocket(multicastPort);
                multicastSock.joinGroup(multicastAddr);
                pack.setAddress(multicastAddr);
                pack.setPort(multicastPort);
                // inizio a inviare linee del file
                
                MulticastMessageSender m = new MulticastMessageSender(f, pack, multicastSock);
                m.start();

                //comunicazione con l'utente
                System.out.println("Il server sta inviando linee, digita \"stop\" per terminare");
                String requested;
                while(true){
                    if(!m.isAlive()){
                        System.out.println("Esecuzione interrotta per errore nell'invio messaggi");
                        break;
                    }
                    BufferedReader br = new BufferedReader(new InputStreamReader(System.in));
                    requested = br.readLine();
                    if(requested.equals("stop")){
                        System.out.println("Terminazione Server");
                        m.stopSending();
                        break;
                    } //if
                }
                //desottoscrizione e chiusura
                richiesta = "MCOFF:" + multicastAddr.getHostAddress() + ":" + multicastPort;
                pack = new DatagramPacket(buf, buf.length, localAddr, discoveryPort);
                pack = Utility.insertStringInPacket(pack, richiesta);
                discoverySocket.send(pack);
                discoverySocket.close();
                multicastSock.close();
            } else {
                System.out.println("Errore di protocollo nella risposta");
                System.exit(1);
            } //if-else

        } catch (Exception ex) {
            System.out.println("Errore: "+ex.getLocalizedMessage());
            if(multicastSock != null && multicastSock.isBound()) multicastSock.close();
            if(discoverySocket.isBound()) discoverySocket.close();
            System.exit(1);
        } //try-catch
    } //main
} //class

